package com.blazinggames.retro.modern.search;

// Java imports
import java.awt.*;
import java.util.Vector;

// Spelchan Imports
import com.blazinggames.urcl.data.Config;
import com.blazinggames.urcl.ui.*;
import com.blazinggames.urcl.ui.sprite.*;
import com.blazinggames.urcl.ui.playfield.*;
import com.blazinggames.urcl.ui.comp.*;

/**
 * Ultimate Retro Project Episode 8: Modern Artillery Battle
 * Copyright (C) 2001, 2003-2004 Blazing Games Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * @author Billy D. Spelchan
 * @version 1.00.011014
*/
public class ArtilleryDisplay 
extends Display
{
// *********************
// ***** CONSTANTS *****
// *********************
    private static final Point DEF_PLACE = new Point (0,0);
    private static final Rectangle DEF_RECT = new Rectangle(0,0,0,0);
    private static final int ANIMATION_FRAMES = 20;

    // Configuration keys
    private static final String BACKDROP_IMAGE = "08BACKDROP";
    private static final String CANNON = "08CANNON";
    // Landscape generation
    private static final String NUM_LANDSCAPES = "08NUMLANDSCAPES";
    private static final String LANDSCAPE = "08LANDSCAPE";
    private static final String TEXTURE = "08LSTEXTURE";

    // Layout
    private static final String MESSAGES = "08MESSAGES";
    private static final String PLAYFIELD = "08PLAYFIELD";
	// LD - large decrement, D - decrement, T - text, I - increment, LI - Large Increment 
    private static final String ANGLE_LD = "08ANGLE_LD";
    private static final String ANGLE_D = "08ANGLE_D";
    private static final String ANGLE_T = "08ANGLE_T";
    private static final String ANGLE_I = "08ANGLE_I";
    private static final String ANGLE_LI = "08ANGLE_LI";
    private static final String POWER_LD = "08POWER_LD";
    private static final String POWER_D = "08POWER_D";
    private static final String POWER_T = "08POWER_T";
    private static final String POWER_I = "08POWER_I";
    private static final String POWER_LI = "08POWER_LI";
    private static final String FIRE = "08FIRE";
//    private static final String SURRENDER = "08SURRENDER";

    // Sound support
    private static final String USE_SOUND = "08USESOUND";
    private static final String SOUND_IMAGE = "08SOUNDIMAGE";
    private static final String SOUND_RECT = "08SOUNDRECT";
    private static final String LAUNCH_SOUND = "08LAUNCH";
    private static final String WIN_SOUND = "08WIN";
    private static final String EXPLODE_SOUND = "08EXPLODE";
    
    // Shared constants (also used by ArtilleryDisplay)
    private static final String NUMPLAYERS = "08NUMPLAYERS";
    private static final String PLAYERNAME = "08PLAYERNAME";
    private static final String PLAYERCOLORS = "08PLAYERCOLOR";


    // Button constants
    private static final int BTN_ANGLE_LD = 0;
    private static final int BTN_ANGLE_D = 1;
    private static final int BTN_ANGLE_I = 2;
    private static final int BTN_ANGLE_LI = 3;
    private static final int BTN_POWER_LD = 4;
    private static final int BTN_POWER_D = 5;
    private static final int BTN_POWER_I = 6;
    private static final int BTN_POWER_LI = 7;
    private static final int BTN_FIRE = 8;
//    private static final int BTN_SURRENDER = 9;
    private static final int TOTAL_BUTTONS = 9;


// *********************
// ***** VARIABLES *****
// *********************

//    Image _background;
    java.util.Random _random;
    boolean _fullPaint;

    // sprites
    private Color _playerColors[][];

    // Playfield
    Playfield _playfield[];
    Rectangle _viewport;
    Point _viewportOffset;
    TileRenderer _blackTile = new TileRenderer(8,8,Color.black);
    TileRenderer _whiteTile = new TileRenderer(8,8,Color.white);
    ImageStripClipped _texture[];
    private int _textureTiles[];
    private int _textureColumns[];

    // Landscape info
    private int _lsMaxHeights[];
    private int _lsColPerGroup[];
    private int _lsSpans[];

    // Sound management
    boolean _useSound;
    boolean _soundToggle;
    
    // Layout and components
    SLabel _messages;
    SLabel _angleText;
    SLabel _powerText;
    TextButton _buttons[];
    
    // Game settings
    private int _landscape;

    private int _min_angle = 5;
    private int _max_angle = 175;
    private int _min_power = 5;
    private int _max_power = 1000;

    // Gameplay
    private int _curPlayer;
    private String _playerNames[];
    private int _playerType[];
    private int _playerAngle[];
    private int _playerPower[];
    //    private ImageStripSprite _playerTanks[];
//    private Sprite _playerTanks[]; 
    private ImageStripSprite _playerTanks[]; 
    private boolean _playerAlive[];
    private int _columns[]; // used to store which column player is in;
    private boolean _gameOver;
    
    // bullet
    private double _bulletX;
    private double _bulletY;
    private double _xVelocity;
    private double _yVelocity;
    private Vector _path = new Vector();
    private boolean _bulletMoving;
    private Point _launchPoint = new Point();
    private Point _lastImpact[];
    private ImageStripSprite _explosion;
    private int _explosionFrame;

    // AI
    private int _aiLastTarget[];
    private int _aiHighs[];
    private int _aiLows[];
    private boolean _aiLastShotAscending[];
    private boolean _aiQuit;


// *****************************************
// ***** METHODS - Initialization aids *****
// *****************************************

    /**
     * prepares all the label fonts as well as loads in the player colors
    */
    private void prepareSprites()
    {
        Container c = _displayHandler.getContainer();
        int n = 8;// _config.getIntProperty(NUMPLAYERS, 2);
        _playerNames = new String[n];
        _playerColors = new Color[n][3];
        _playerAngle = new int[n];
        _playerPower = new int[n];
        _playerType = new int[n];
        _playerTanks = new ImageStripSprite[n];
        _playerAlive = new boolean[n];

    	_lastImpact = new Point[n];
    
        _aiLastTarget = new int[n];
        _aiHighs = new int[n];
        _aiLows = new int[n];
        _aiLastShotAscending = new boolean[n];
        
        _res.addImage(CANNON, _config.getStringProperty(CANNON, "res/artillery/cannon.gif"));
        String tempName;
        for (int cntr = 0; cntr < n; ++cntr)
        {
            _playerNames[cntr] = _config.getStringProperty(
                        PLAYERNAME + "[" + cntr + "]", "Player " + (cntr + 1));
            _playerColors[cntr][1] = _config.getColorProperty(PLAYERCOLORS + "[" + cntr + "]", Color.white);
            _playerColors[cntr][2] = _playerColors[cntr][1].brighter();
            _playerColors[cntr][0] = _playerColors[cntr][1].darker();
            _playerAngle[cntr] = _min_angle;
            _playerPower[cntr] = _min_power;
            _playerTanks[cntr] = new ImageStripSprite(c, _res.getImage(CANNON), 8, 8);
            _playerTanks[cntr].setFrame(20*cntr+1);
            _playerAlive[cntr] = false;
    
        }
    
        _explosion = new ImageStripSprite(c, _res.getImage(CANNON), 8, 8);
    }


    /**
     * Prepares the hall, which is simply a playfield so it can easily be scrolled.
    */
    private void preparePlayfield()
    {
        Container c = _displayHandler.getContainer();
	
	Rectangle rect = _config.getRectangleProperty(PLAYFIELD, DEF_RECT);
	_viewport = new Rectangle(0,0,rect.width, rect.height);
	_viewportOffset = new Point(rect.x, rect.y);
	
	int n = _config.getIntProperty(NUM_LANDSCAPES, 0);
	if (n < 1) 
	    n = 1;
	_playfield = new Playfield[n];
	_texture = new ImageStripClipped[n];
	_lsMaxHeights = new int[n];
	_lsColPerGroup = new int[n];
	_lsSpans = new int[n];
	_textureTiles = new int[n];
	_textureColumns = new int[n];
	String args[];

	for (int cntr = 0; cntr < n; ++cntr) {
	    _playfield[cntr] = new Playfield();
	    _playfield[cntr].setTileset(4,8,8);
	    _playfield[cntr].setTileRenderer(0, _blackTile);
	    _playfield[cntr].setTileRenderer(2, _blackTile); // used as marker

	    args = _config.parseString(_config.getStringProperty(TEXTURE+"["+cntr+"]","x"), ',');
	    if (args.length != 3)
	    {
		_playfield[cntr].setTileRenderer(1, _whiteTile);
		_textureTiles[cntr] = 1;
		_textureColumns[cntr] = 1;
	    }
	    else
	    {
		_textureTiles[cntr] = _config.parseInt(args[1], 1);
		_textureColumns[cntr] = _config.parseInt(args[2], 1);
		_res.addImage(args[0], args[0]);
		_texture[cntr] = new ImageStripClipped(c, _res.getImage(args[0]), 8, 8);
		_playfield[cntr].setTileRenderer(1, _texture[cntr]);
	    }

	    args = _config.parseString(_config.getStringProperty(LANDSCAPE+"["+cntr+"]", "x"), ',');
	    if (args.length != 3) {
		_lsMaxHeights[cntr] = 40;
		_lsColPerGroup[cntr] = 1;
		_lsSpans[cntr] = 5;
	    }
	    else {
		_lsMaxHeights[cntr] = _config.parseInt(args[0], 40);
		_lsColPerGroup[cntr] = _config.parseInt(args[1], 1);
		_lsSpans[cntr] = _config.parseInt(args[2], 4);
	    }
	}
    }

    private void prepareSounds()
    {
        _useSound = (_config.getIntProperty(USE_SOUND, 0) == 1);
        _res.addImage(SOUND_IMAGE, _config.getStringProperty(SOUND_IMAGE, "sound.gif"));
        Rectangle r = _config.getRectangleProperty(SOUND_RECT, DEF_RECT);
    	_soundToggle = false;
        
        if (_useSound)
        {
            _res.addAudioClip(LAUNCH_SOUND, _config.getStringProperty(LAUNCH_SOUND, "silent.au"));
            _res.addAudioClip(WIN_SOUND, _config.getStringProperty(WIN_SOUND, "silent.au"));
            _res.addAudioClip(EXPLODE_SOUND, _config.getStringProperty(EXPLODE_SOUND, "silent.au"));
        }
    }

    /**
     * processes the layout
    */
    public void prepareLayout()
    {
        Container c = _displayHandler.getContainer();

	_messages = new SLabel(c, "Place messages here", 
			       _config.getRectangleProperty(MESSAGES, DEF_RECT));
	_angleText = new SLabel(c, "Ang", _config.getRectangleProperty(ANGLE_T, DEF_RECT));
	_powerText = new SLabel(c, "Pow", _config.getRectangleProperty(POWER_T, DEF_RECT));
        _buttons = new TextButton[TOTAL_BUTTONS];

	_buttons[BTN_ANGLE_LD] = new TextButton(c, "-5", _config.getRectangleProperty(ANGLE_LD, DEF_RECT));
	_buttons[BTN_ANGLE_D] = new TextButton(c, "-1", _config.getRectangleProperty(ANGLE_D, DEF_RECT));
	_buttons[BTN_ANGLE_I] = new TextButton(c, "+1", _config.getRectangleProperty(ANGLE_I, DEF_RECT));
	_buttons[BTN_ANGLE_LI] = new TextButton(c, "+5", _config.getRectangleProperty(ANGLE_LI, DEF_RECT));
	_buttons[BTN_POWER_LD] = new TextButton(c, "-10", _config.getRectangleProperty(POWER_LD, DEF_RECT));
	_buttons[BTN_POWER_D] = new TextButton(c, "-1", _config.getRectangleProperty(POWER_D, DEF_RECT));
	_buttons[BTN_POWER_I] = new TextButton(c, "+1", _config.getRectangleProperty(POWER_I, DEF_RECT));
	_buttons[BTN_POWER_LI] = new TextButton(c, "+10", _config.getRectangleProperty(POWER_LI, DEF_RECT));
	_buttons[BTN_FIRE] = new TextButton(c, "FIRE", _config.getRectangleProperty(FIRE, DEF_RECT));
//	_buttons[BTN_SURRENDER] = new TextButton(c, "SURRENDER", _config.getRectangleProperty(SURRENDER, DEF_RECT));

	// TEST CODE
	//	_landscape = 0;
	//	generateLandscape();
    }


// ***********************************
// ***** METHODS -  From Display *****
// ***********************************

    /**
     * Initializes the game (using private helper methods) by getting necessary
     * resources and setting up configurable variables.
     *
     * @param dh the controlling applet/application
    */
    public boolean initialize(DisplayHandler dh)
    {
        super.initialize(dh);
        _random = new java.util.Random();
        dh.displayEvent(this, "LOADING 20");
//        _res.addImage(BACKDROP_IMAGE, _config.getStringProperty(BACKDROP_IMAGE, "backdrop.jpg"));
//        _background = _res.getImage(BACKDROP_IMAGE);
        dh.displayEvent(this, "LOADING 40");
        prepareSprites();
        dh.displayEvent(this, "LOADING 60");
        prepareSounds();
        preparePlayfield();
        dh.displayEvent(this, "LOADING 80");
        prepareLayout();
        dh.displayEvent(this, "LOADING 100");

        return true;
    }

    public boolean begin()
    {
	_fullPaint = true;
	return super.begin();
    }


    /**
     * Handles mouse movement. We may need to update the current guess variable
     * and if the game is over, the quit button needs messages.
     *
     * @param x X coordinate of mouse pointer
     * @param y Y coordinate of mouse pointer
     * @param b true if mouse button down
    */
    public void mouse_move(int x, int y, boolean b)
    {
        boolean repaint = false;
        
	for (int cntr = 0; cntr < TOTAL_BUTTONS; ++cntr)
	{
	    repaint |= _buttons[cntr].mouse_move(x, y, b);
	}

        if (repaint)
            _displayHandler.refreshDisplay(this);
    }
    
    /**
     * Handles the mouse being clicked. If over the hallway (playfield) then
     * player may be starting to make a guess. If the game is over then we
     * have to pass messages to the quit button.
     *
     * @param x X coordinate of mouse pointer
     * @param y Y coordinate of mouse pointer
    */
    public void mouse_down(int x, int y)
    {
        boolean repaint = false;

	for (int cntr = 0; cntr < TOTAL_BUTTONS; ++cntr)
	{
	    repaint |= _buttons[cntr].mouse_down(x, y);
	}

        if (repaint)
            _displayHandler.refreshDisplay(this);
    }
    
    /**
     * Handles the mouse button being released. This sends messages to the
     * scrollbar and (if game over) quit button components as well as handling
     * clicking on a door in the hallway playfield.
     *
     * @param x X coordinate of mouse pointer
     * @param y Y coordinate of mouse pointer
    */
    public void mouse_up(int x, int y)
    {
        boolean repaint = true;

        if (_gameOver)
        {
            _displayHandler.displayEvent(this, "ARTILLERY_TITLE");
            return;
        }

        for (int cntr = 0; cntr < TOTAL_BUTTONS; ++cntr)
        {
            if (_buttons[cntr].mouse_up(x, y))
            {
                repaint = true;
                handleButton(cntr);
            }
        }
    
        if (repaint)
            _displayHandler.refreshDisplay(this);
    }
    
    /**
     * Ticks occur at a specified interval (100ms in my case).
     *
     * @param ms milliseconds between this click and the last one
    */
    public void tick(int ms)
    {
        boolean repaint = false;
        int cntr;
    
        for (cntr = 0; cntr < TOTAL_BUTTONS; ++cntr)
        {
            repaint |= _buttons[cntr].tick(ms);
        }
    
        if (_bulletMoving)
        {
            do {} while (moveBullet() == false);
            repaint = true;
        }
    
        if (_explosionFrame < ANIMATION_FRAMES)
        {
            repaint = true;
            _explosion.setFrame(180+_explosionFrame);
          ++_explosionFrame;
            if (_explosionFrame == ANIMATION_FRAMES)
            {
                if (_lastImpact[_curPlayer].x >= 0) // make sure hit tile is in playfield
                {
                    for (cntr = _lastImpact[_curPlayer].y - 1; cntr > -1; --cntr)
                        _playfield[_landscape].setTile( _lastImpact[_curPlayer].x, cntr + 1, _playfield[_landscape].getTile(_lastImpact[_curPlayer].x, cntr), _playfield[_landscape].getAttribute(_lastImpact[_curPlayer].x, cntr));
                    _playfield[_landscape].setTile(_lastImpact[_curPlayer].x, 0, 0, 0);

                    // see if have to adjust sprite and if so do so
                    for (cntr = 0; cntr < _playerTanks.length; ++cntr)
                    {
                        if (_columns[cntr] == _lastImpact[_curPlayer].x)
                        {
                            Point p = _playerTanks[cntr].getLocation();
                            _playerTanks[cntr].move(p.x, p.y + 8);
                        }
                    }
                }
                _explosion.setVisible(false);
                _path.removeAllElements();
                if (_gameOver != true)
                    nextPlayer();//_fullPaint = true;
            }
        }
    
        if (repaint)
        {
            _displayHandler.refreshDisplay(this);
        }
    }

    /**
     * draws the display screen to the indicated graphics context.
     *
     * @param g Graphics context to draw to
     * @param comp Component for awt calls
    */
    public synchronized void render(Graphics g, Component comp)
    {
        int cntr;
        Dimension d = comp.getSize();

        if (_fullPaint)
        {
            _fullPaint = false;
            _playfield[_landscape].render(g, comp, _viewportOffset.x, _viewportOffset.y, _viewport);
    
            // player tanks
            for (cntr = 0; cntr < _playerTanks.length; ++cntr)
            {
                _playerTanks[cntr].setFrame(cntr*20 + _playerAngle[cntr] / 10);
                _playerTanks[cntr].render(g, comp, _viewportOffset);
            }
        }

        _playerTanks[_curPlayer].setFrame(20*_curPlayer + _playerAngle[_curPlayer] / 10);
        _playerTanks[_curPlayer].render(g, comp, _viewportOffset);

        // interface elements
        g.setColor(_playerColors[_curPlayer][2]);
        int fillStartY = _viewportOffset.y + _viewport.height;
        g.fillRect(0,fillStartY, d.width, d.height - fillStartY);
        for (cntr = 0; cntr < TOTAL_BUTTONS; ++cntr)
        {
            _buttons[cntr].render(g, comp);
        }
        _messages.render(g, comp);
        _angleText.render(g, comp);
        _powerText.render(g, comp);
    
        // current shot
        if (_path.size() < 2)
            return;
        Point pStart = (Point)_path.elementAt(0);
        Point pEnd;
        g.setColor(_playerColors[_curPlayer][1]);
        for (cntr = 1; cntr < _path.size(); ++cntr)
        {
            pEnd = (Point)_path.elementAt(cntr);
            g.drawLine(pStart.x, pStart.y, pEnd.x, pEnd.y);
            pStart = pEnd;
        }
        _explosion.render(g, comp);
    }
    
    
    /**
     * If sound is enabled gets the resource manager to play the indicated sound.
     *
     * @param key sound to play
    */
    private void playSound(String key)
    {
        if ( (_useSound) && (_soundToggle) )
        {
            _res.playAudio(key);
        }
    }


// ***********************************************
// ***** METHODS - User Interface Management *****
// ***********************************************

    private void setButtonState(int player, boolean visible)
    {
        Container c = _displayHandler.getContainer();

        for (int cntr = 0; cntr < TOTAL_BUTTONS; ++cntr)
        {
            _buttons[cntr].setNormalColors(_playerColors[player][0], Color.black);
            _buttons[cntr].setOverColors(_playerColors[player][1], Color.black);
            _buttons[cntr].setSelectedColors(_playerColors[player][2], Color.black);
            _buttons[cntr].setVisible(visible);
        }
        _messages.setBackground(_playerColors[player][1]);
        _messages.setFontColor(Color.black);
        _angleText.setText(c, "Angle " + _playerAngle[player]);
    	_powerText.setText(c, "Power " + _playerPower[player]);

//        _fullPaint = true;
    }

    private void setMessage(String msg)
    {
        _messages.setText(_displayHandler.getContainer(), msg);
    }
    
    private void handleButton(int id)
    {
        switch (id)
        {
        case BTN_ANGLE_LD:
            _playerAngle[_curPlayer] = Math.max(_min_angle, _playerAngle[_curPlayer] - 5);
            break;
    
        case BTN_ANGLE_D:
            _playerAngle[_curPlayer] = Math.max(_min_angle, _playerAngle[_curPlayer] - 1);
            break;
    
        case BTN_ANGLE_I:
            _playerAngle[_curPlayer] = Math.min(_max_angle, _playerAngle[_curPlayer] + 1);
            break;
    
        case BTN_ANGLE_LI:
            _playerAngle[_curPlayer] = Math.min(_max_angle, _playerAngle[_curPlayer] + 5);
            break;
    
        case BTN_POWER_LD:
            _playerPower[_curPlayer] = Math.max(_min_power, _playerPower[_curPlayer] - 10);
            break;
    
        case BTN_POWER_D:
            _playerPower[_curPlayer] = Math.max(_min_power, _playerPower[_curPlayer] - 1);
            break;
    
        case BTN_POWER_I:
            _playerPower[_curPlayer] = Math.min(_max_power, _playerPower[_curPlayer] + 1);
            break;
    
        case BTN_POWER_LI:
            _playerPower[_curPlayer] = Math.min(_max_power, _playerPower[_curPlayer] + 10);
            break;
    
        case BTN_FIRE:
            launchBullet();
            break;
/*    
        case BTN_SURRENDER:
            // TEST START
            _displayHandler.displayEvent(this, "ARTILLERY_TITLE");
            _landscape = (_landscape + 1) % 3;
            generateLandscape();
            // TEST END
            break;
*/
        }
        Container c = _displayHandler.getContainer();
        _angleText.setText(c, "Angle " + _playerAngle[_curPlayer]);
        _powerText.setText(c, "Power " + _playerPower[_curPlayer]);
    }


// ***********************************
// ***** METHODS - Game Handling *****
// ***********************************

    public int rand(int low, int high)
    {
        int r = high - low + 1;
        if (r == 1)
            return low;
        return (((_random.nextInt() & Integer.MAX_VALUE) % r) + low);
    }

    private void generateLandscape()
    {
        _playfield[_landscape].setPlayfield(80, 45);
        int lastHeight = rand(1, _lsMaxHeights[_landscape]);
        int lowRange, highRange, curHeight, groupHeight, tileAttr;
        int cpg = _lsColPerGroup[_landscape];
        for (int cntrX = 0; cntrX < 80; cntrX += cpg) 
        {
            lowRange = Math.max(1, lastHeight - _lsSpans[_landscape]);
            highRange = Math.min(_lsMaxHeights[_landscape], lastHeight + _lsSpans[_landscape]);
            groupHeight = rand(lowRange, highRange);
            for (int cntrG = 0; cntrG < cpg; ++cntrG) 
            {
                tileAttr = (cntrX + cntrG) % _textureColumns[_landscape];
                curHeight = lastHeight - (lastHeight - groupHeight) * (cntrG+1) / cpg;
                for (int cntrY = (45 - curHeight); cntrY < 45; ++cntrY) 
                {
                    _playfield[_landscape].setTile(cntrX + cntrG, cntrY, 1, tileAttr);
                    tileAttr = (tileAttr + _textureColumns[_landscape]) % _textureTiles[_landscape];
                }
            }
            lastHeight = groupHeight;
        }
    }

    public void setPlayerType(int id, int type)
    {
        _playerType[id] = type;
    }

    public void setLandscapeType(int id)
    {
        _landscape = id;
    }
    
    public void setSound(boolean play)
    {
        _soundToggle = play;
    }
    
    public void setAIQuit(boolean quit)
    {
        _aiQuit = quit;
    }
    
    public void startGame()
    {
        _columns = new int[_playerTanks.length];
        int row, cntr, cntrCheck, cntrRow;
        boolean valid;
        
        _gameOver = false;
        _bulletMoving = false;
        generateLandscape();
        for (cntr = 0; cntr < _playerTanks.length; ++cntr)
        {
            _playerAlive[cntr] = _playerType[cntr] > 0;
            _aiLastTarget[cntr] = -1;
            _aiHighs[cntr] = 0;
            _aiLows[cntr] = 0;
            do
            {
                _columns[cntr] = rand(1, 79);
                valid = true;
                for (cntrCheck = 0; cntrCheck < cntr; ++cntrCheck)
                {
                    if ( (_columns[cntr] >= (_columns[cntrCheck] - 1)) &&
                    (_columns[cntr] <= (_columns[cntrCheck] + 1)) )
                    valid = false;
                }
            }
            while (valid == false);

            row = 44;
            for (cntrRow = 0; cntrRow < 45; ++cntrRow)
            {
                if (_playfield[_landscape].getTile(_columns[cntr], cntrRow) == 1)
                {
                    row = cntrRow - 1;
                    break;
                }
            }
            _playerAngle[cntr] = 90;
            _playerPower[cntr] = 100;
            _playerTanks[cntr].move(_columns[cntr] * 8, row * 8);
            _playerTanks[cntr].setVisible(_playerAlive[cntr]);
            _playfield[_landscape].setTile(_columns[cntr], row, 2, 1);
        }

        _curPlayer = 7;
        nextPlayer();
    
        _explosionFrame = ANIMATION_FRAMES; // set to end so no starting delay due to anim.
        _explosion.setVisible(false);
    }

    /**
     * Checks to see if the game is over. If it isn't finds out who the next
     * player is and sets the game up to use that player.
     */
    private void nextPlayer()
    {
        Container c = _displayHandler.getContainer();
        int tankCount = 0;
        int lastCounted = 0;
        for (int cntr = 0; cntr < 8; ++cntr)
        {
            if (_playerAlive[cntr])
            {
                ++tankCount;
                lastCounted = cntr;
            }
        }
        if (tankCount == 1)
        {
            playSound(WIN_SOUND);
            _messages.setText(c, _playerNames[lastCounted] + " won! Click to return to title");
            _gameOver = true;
            _curPlayer = lastCounted;
            setButtonState(lastCounted, false);
            return;
        }
        else if (tankCount == 0)
        {
            _messages.setText(c, "NO WINNER! Click to return to title");
            _gameOver = true;
            setButtonState(_curPlayer, false);
            return;
        }
        
        // find next player
        boolean valid = false;
        do
        {
            ++_curPlayer;
            if (_curPlayer > 7)
            _curPlayer = 0;
            valid = _playerAlive[_curPlayer];
        }
        while (valid == false);
        setMessage(_playerNames[_curPlayer] +"s Turn...");
        if (_playerType[_curPlayer] == 1)
            setButtonState(_curPlayer, true);
        else
            computerTurn();
        _fullPaint = true;
    }

    private void computerTurn()
    {
        if (_playerType[_curPlayer] == 2) // is dumb ai?
        {
            _playerAngle[_curPlayer] = rand(5, 175);
            _playerPower[_curPlayer] = rand(50, 500);
            launchBullet();
        }
        else // its using the smart ai 
        {
            boolean needTarget = (_aiLastTarget[_curPlayer] == -1);
            if (needTarget == false)
                needTarget = ! _playerAlive[_aiLastTarget[_curPlayer]];
            if (needTarget)
            {
                // Pick new target
                int target;
                do
                {
                    target = rand(0, 7);
                }
                while ( (target == _curPlayer) || (_playerAlive[target] == false) );
                _aiLastTarget[_curPlayer] = target;
                _aiHighs[_curPlayer] = 0;
                _aiLows[_curPlayer] = 0;
                _displayHandler.log("AI " + _curPlayer + " targetting " + target);
                if (_columns[_curPlayer] < _columns[_aiLastTarget[_curPlayer]])
                    _playerAngle[_curPlayer] = rand(5, 89);
                else
                    _playerAngle[_curPlayer] = rand(91, 175);
                _playerPower[_curPlayer] = rand(50, 200);                    
            }
            else
            {
                // Adjust shot
                int targX = _columns[_aiLastTarget[_curPlayer]];
                int curPow = _playerPower[_curPlayer];
                if (_columns[_curPlayer] < targX)
                {	// target to right
                    if (_aiHighs[_curPlayer] > 0)
                    {
                        if (_lastImpact[_curPlayer].x < targX)
                            _aiLows[_curPlayer] = curPow;
                        else
                            _aiHighs[_curPlayer] = curPow;
                        _playerPower[_curPlayer] = rand(_aiLows[_curPlayer], _aiHighs[_curPlayer]);
                    }
                    else if (_lastImpact[_curPlayer].x < targX)
                    {
                        if (_aiLastShotAscending[_curPlayer])
                            _playerAngle[_curPlayer] = rand(_playerAngle[_curPlayer] ,89);
                        else
                            _playerPower[_curPlayer] = rand(curPow+1, curPow+100);
                    }
                    else
                    {
                        _aiHighs[_curPlayer] = curPow;
                        _playerPower[_curPlayer] = rand(_aiLows[_curPlayer], _aiHighs[_curPlayer]);
                    }
                }
                else
                {	// target to left
                    if (_aiHighs[_curPlayer] > 0)
                    {
                        if (_lastImpact[_curPlayer].x > targX)
                            _aiLows[_curPlayer] = curPow;
                        else
                            _aiHighs[_curPlayer] = curPow;
                        _playerPower[_curPlayer] = rand(_aiLows[_curPlayer], _aiHighs[_curPlayer]);
                    }
                    else if (_lastImpact[_curPlayer].x > targX)
                    {
                        if (_aiLastShotAscending[_curPlayer])
                            _playerAngle[_curPlayer] = rand(91, _playerAngle[_curPlayer]);
                        else
                            _playerPower[_curPlayer] = rand(curPow+1, curPow+100);
                    }
                    else
                    {
                        _aiHighs[_curPlayer] = curPow;
                        _playerPower[_curPlayer] = rand(_aiLows[_curPlayer], _aiHighs[_curPlayer]);
                    }
//                    _playerAngle[_curPlayer] = rand(91, 175);
                }
            }

/*            
            _displayHandler.log("No advanced ai yet so shooting self");
            _playerAngle[_curPlayer] = 90;
            _playerPower[_curPlayer] = 50;
*/
            launchBullet();
        }
    }

    private void launchBullet()
    {
        _bulletX = _playerTanks[_curPlayer].getX() * 10 + 40;
        _bulletY = (360 - _playerTanks[_curPlayer].getY()) * 10 - 40;
        double rad = (double)_playerAngle[_curPlayer] * Math.PI / 180;
        _xVelocity = Math.cos(rad) * (double)_playerPower[_curPlayer];
        _yVelocity = Math.sin(rad) * (double)_playerPower[_curPlayer];
        _path.removeAllElements();
        _bulletMoving = true;
        _path.addElement(new Point((int)(_bulletX / 10) + _viewportOffset.x, 
                    (int)(360 - _bulletY/10) + _viewportOffset.y));
        _launchPoint.setLocation(_playerTanks[_curPlayer].getX() / 8, 
                    _playerTanks[_curPlayer].getY() / 8);
        setButtonState(_curPlayer, false);
        setMessage(_playerNames[_curPlayer] + " Firing at " + _playerAngle[_curPlayer] + " Power " + _playerPower[_curPlayer]);
        playSound(LAUNCH_SOUND);
    }

    /**
        returns false if bullet out of bounds
    */
    private boolean moveBullet()
    {
        int x1 = (int)(_bulletX / 10);
        int y1 = (int)(_bulletY / 10);
        _bulletX += (_xVelocity / 10);
        _bulletY += (_yVelocity / 10);
        _yVelocity -= 0.98;
        int x2 = (int)(_bulletX / 10);
        int y2 = (int)(_bulletY / 10);
    
//        _displayHandler.log("Bullet at " + _bulletX + ", " + _bulletY);
        if (_bulletY <= 0)
            _bulletMoving = false;
        if (y1 > 360) // remember different coordinate systems
        {
            if (y2 <= 360)
            {
                _path.addElement(new Point(x1 + _viewportOffset.x, 360 - y1 + _viewportOffset.y));
                _path.addElement(new Point(x2 + _viewportOffset.x, 360 - y2 + _viewportOffset.y));
            }
            if ((x2 > 0) && (x2 < 640))
                return false;
        }
        else
        {
            _path.addElement(new Point(x2 + _viewportOffset.x, 360 - y2 + _viewportOffset.y));
//            _path.addElement(new Point((int)(_bulletX / 10) + _viewportOffset.x, 
//                    (int)(360 - _bulletY/10) + _viewportOffset.y));
        }
    
        int lastTileY = 44 - y1 / 8;
        int xSpan = x2 - x1;
        int ySpan = y2 - y1;
        int cntr, curTileY;
    
        if (xSpan == 0)
            xSpan = 1;
        double m = (double)ySpan / (double)xSpan;
        double y;
        Point impact = null;
    
        if (x1 > x2)
        {
            for (cntr = x1; cntr >= x2; --cntr)
            {
                y = m * (cntr - x1) + y1;
                curTileY = 44 - (int)(y / 8);
                impact = checkRowImpact(cntr / 8, lastTileY, curTileY);
                if ( (impact != null) && ((_yVelocity < 0) || (_launchPoint.equals(impact) == false) ) )
                    break;
                lastTileY = curTileY;
            }
        }
        else
        {
            for (cntr = x1; cntr <= x2; ++cntr)
            {
                y = m * (cntr - x1) + y1;
                curTileY = 44 - (int)(y / 8);
                impact = checkRowImpact(cntr / 8, lastTileY, curTileY);
                if ( (impact != null) && (_launchPoint.equals(impact) == false) )
                    break;
                lastTileY = curTileY;
            }
        }
    
        if ( (impact != null) && (_launchPoint.equals(impact) == false) )
        {
            _bulletMoving = false;
            _aiLastShotAscending[_curPlayer] = _yVelocity > 0;
            
            // we are determining self hits as tile under tank as bullet has to pass through tank
            if ((impact.x == _launchPoint.x) && (impact.y == (_launchPoint.y + 1)))
            {
                _messages.setText(_displayHandler.getContainer(), "PLAYER SHOT SELF!!!");
                --impact.y;
            }
            
            if (impact.x >= 0)
            {
                playSound(EXPLODE_SOUND);
                int tile = _playfield[_landscape].getTile(impact.x, impact.y);
                if (tile == 2)
                {
                    setMessage(_playerNames[_curPlayer] + " HIT TANK!");
                    _displayHandler.log("***** HIT TANK!!! *****");
                    int numHumansLeft = 0;
                    for (int cntrPlayer = 0; cntrPlayer < 8; ++cntrPlayer)
                    {
                        if (_columns[cntrPlayer] == impact.x)
                        {
                            _playerAlive[cntrPlayer] = false;
                            _playerTanks[cntrPlayer].setVisible(false);
                        }
                        else
                        {
                            if ( (_playerAlive[cntrPlayer] == true) && (_playerType[cntrPlayer] == 1))
                                ++numHumansLeft;
                        }
                    }
                    if ( (_aiQuit == true) && (numHumansLeft == 0) )
                    {
                        setMessage("Only computer players left - click to continue");
                        _gameOver = true;
                    }
                }
                else
                {
                    setMessage(_playerNames[_curPlayer] + " missed!");
                    _displayHandler.log("Hit tile " + impact.x + "," + impact.y);
                }
                _explosion.move(impact.x * 8 + _viewportOffset.x, impact.y * 8 + _viewportOffset.y);
                _explosion.setVisible(true);
                _explosionFrame = 0;
                _lastImpact[_curPlayer] = impact;
            }
            else // out of bounds shot
            {
                setMessage(_playerNames[_curPlayer] + " shot out of bounds!!!");
                _explosion.setVisible(false);
                _explosionFrame = 0;
                _lastImpact[_curPlayer] = impact;
            }
        }
        return true;
    }

    private Point checkRowImpact(int x, int y1, int y2)
    {
	if ( (x < 0) || (x > 79) )
	    return new Point(-1, -1); // indicate dead point
	int cntr, impY = 0;
	boolean impacted = false;
//	_displayHandler.log("Checking row " + x + " columns " + y1 + " to " + y2);
	if (y1 < y2)
	{
	    for (cntr = y1; cntr <= y2; ++cntr)
	    {
		if (_playfield[_landscape].getTile(x, cntr) > 0)
		{
		    impY = cntr;
		    impacted = true;
		    break;
		}
	    }
	}
	else
	{
	    for (cntr = y1; cntr >= y2; --cntr)
	    {
		if (_playfield[_landscape].getTile(x, cntr) > 0)
		{
		    impY = cntr;
		    impacted = true;
		    break;
		}
	    }
	}
	if (impacted)
	    return new Point(x, impY);
	return null;
    }
}
