package com.blazinggames.urcl.ui.comp;

// Java imports
import java.awt.*;

// Spelchan imports
import com.blazinggames.urcl.ui.sprite.*;

/**
* Your typical radio component. To be usefull, it has to be used with
* a RadioContainer class.
*
* @author Billy D. Spelchan
* @version 1.10.031028
*/
public class SRadioButton
extends Comp
{
    // *********************
    // ***** CONSTANTS *****
    // *********************

    private static final int GAP_SIZE = 5;


    // *********************
    // ***** VARIABLES *****
    // *********************

    protected int _state;
    protected Color _foreground;
    protected Color _background;


    // ************************
    // ***** CONSTRUCTORS *****
    // ************************

    public SRadioButton(Component c, String s, int x, int y)
    {
        super();
        _sprite = new TextSprite(c, s, x, y);
        _sprite.setSolid(false);
        setBackground(Color.white);
        setForeground(Color.black);
        setBounds(x,y, _sprite.getHeight() + GAP_SIZE + _sprite.getWidth(), _sprite.getHeight());
        _state = 0;
    }


    // *******************************
    // ***** METHODS - From Comp *****
    // *******************************

    /**
        * Sets if a component is active or not. A non-active component generally
        * looks different and doesn't respond to events.
        *
        * @param b true if component is to be active, false otherwise
        */
    public void setActive(boolean b)
    {
        super.setActive(b);

        if ( (b) && (_state < 0) )
            _state = 0;
        else if (b == false)
            _state = -1;
    }

    /**
        * Draws the component.
        *
        * @param g Graphics Context to draw into
        * @param c Component class that some awt commands need
        */
    public void render(Graphics g, Component c)
    {
        _changed = false;
        int h = _sprite.getHeight();
        g.setColor(_foreground);
        g.fillOval(_bounds.x, _bounds.y, h, h);
        g.setColor(_background);
        g.fillOval(_bounds.x + 2, _bounds.y + 2, h - 4, h - 4);
        if (_state == 1) {
            g.setColor(_foreground);
            g.fillOval(_bounds.x + 4, _bounds.y + 4, h - 8, h - 8); 
        }

        _sprite.move(_bounds.x + GAP_SIZE + _sprite.getHeight(), _bounds.y);
        _sprite.setVisible(_isVisible);
        _sprite.setBounding(_isSelected);
        _sprite.render(g, c);
        _sprite.clean();
    }


    // *****************************************************
    // ***** METHODS - Information getting and setting *****
    // *****************************************************

    /** 
        * Changes the state of the toggle to the indicated state.
        *
        * @param n new state for the component
        */
    public void setState(int n)
    {
        if (_isActive == false)
            return;
        if (n >= 2)
        {
            _state = 0;
        }
        else
        {
            _state = n;
        }
    }

    /**
        * Goes to the next state in the toggle sequence
        */
    public void nextState()
    {
        setState(_state+1);
    }

    /**
        * @return Current state of the toggle (-1 if disabled)
        */
    public int getState()
    {
        return _state;
    }

    public void setForeground(Color c)
    {
        _foreground = c;
        ((TextSprite)_sprite).setFontColor(c);
    }

    public void setBackground(Color c)
    {
        _background = c;
    }

    public void setText(Component c, String s)
    {
        ((TextSprite)_sprite).setText(c, s);
    }

    public void setFont(Component c, Font f)
    {
        ((TextSprite)_sprite).setFont(c, f);
    }

    public void setFormat(int n)
    {
        ((TextSprite)_sprite).setFormat(n);
    }

}

